FORGET SDCARD.fth
TACHYON
[~
{ SD CARD TOOLKIT }    
pub SDCARD.fth    ." SD CARD Toolkit - 131114.0000 " ;

{
DESCRIPTION
These are the general routines used to initialize and access the SD card and can also be used to load and save blocks directly. There are various tools plus the primitives which allow you to directly interact with the card and it's command set. This module forms the foundation on which other modules for virtual memory and file system relies.

NOTES:
SOFT TIMER #7 is used for timeouts
 
NOTE: The card insertion method used here in CARD? is based on that mentioned in the official SD MEMORY CARD SPECIFICATIONS which detects the pullup that exists on the chip select of a memory card itself. This means that a "safety" pull-up is never required and the card detect will not work if one is used.


CHANGELOG:
131114        - tidied and trimmed source
131111        - Standardized for hardware header support
            - Make BUFFERS fully available for 4 sectors (multiple files open)
131105        - Increased timeout from 5,000 to 10,000
131104        - Added in basic virtual memory interface words 
130723        - Restructured the code, optimized SDPINS etc. inserted datasheet images
130717        - Tidied up source
130303        - Changed constant names to =constant rather than #constant as numeric preprocessor clashes
121226        - Removed redundant [SPIO] references which were also slowing down access 
121013        - Modify to use BUFFERS (VM image)
120917        - Added SPINNERET
120909:2250     - Added CARD detect redefinition vecotr "ucard" 
120909:1000        - Added extra pin definitions for different boards
120903:1400        - Added pin redirection

Module Information
NAMES:  $5B46...7496 for 6480 (0608 bytes added)
CODE:   $0000...3766 for 7702 (1818 bytes added)
CALLS:  0423 vectors free
RAM:    9184 bytes free

TESTS:
Reading from the card:
4K LOAD takes 18.76ms
512 byte LOAD takes 2.55ms
}

{ NOTE: Pin defintions may be made at compile time or at runtime using SDPINS
If a header file is compiled beforehand then it should include these pin names as shown.
The card detect is via the SDCS which detects a card's "pullup" (do not use external pullups here!!!)
}
IFNDEF &SDCS
#P26    |<     == &SDCS            \ SDCARD CS
#P27    |<     == &SDDO            \ Data out from SDCARD
#P28    |<     == &SDCK            \ SDCARD clocks (Shared with SCL)
#P29    |<     == &SDDI            \ Data to SDCARD (Shared with SDA)
}

#512        == BLKSIZ

\ SD Tokens
$FE         == =dtk            \ data token for single block read/write
\ $FC         == =dtk1            \ data token for 1st block read/write
\ $FD         == =dtke            \ data token for last block read/write


TABLE sdinfo        #176 ALLOT
sdinfo         ORG            \ Use Tachyon reserved buffer space for SD card (normally at $7500)
#64            DS cid
#64            DS csd
#16            DS extbuf

#4            DS sdpins
#4            DS sdsize
#4            DS sdwr
#4            DS sdrd
#4            DS ocr
#2            DS ucard        \ user vector for card detect
#2            DS sdbusy
#1            DS crc
#1            DS card?        \ card detect transition memory
#1            DS blklen
#1            DS sdhc

1            DS filesel        \ currently selected file 0..3

BYTE wrflg

\ Current sector in SDBUFs
TABLE sector 4 4 * ALLOT
\ Sector CRCs
TABLE sectcrc 4 4 * ALLOT


\ Set the active file channel
pub FILE ( index -- ) 
     filesel C!
     ;

pub @FILE ( addr -- addr+off )
     filesel C@ 2* 2* +
     ;
\ Return with address of currently selected sector buffer
pub SDBUF ( -- buf )
filesel C@ 9 SHL BUFFERS +
;

\ Load default SD pins at compile time ( ce-miso-mosi-clk -- )
&SDCS >| #24 SHL &SDDO >| 16 SHL OR &SDDI >| 8 SHL OR &SDCK >| OR sdpins !



\ Detect SD card presence - the CS line must not have a pullup on it (redundant and undesirable).
pub CARD?    ( -- flg ; Is SD card present? )
     ucard W@ ?DUP IF JUMP THEN        \ use alternative card detect method if selected
    &SDCS OUTCLR                \ Force CS line low
    &SDCS INPUTS                \ Float CS
    &SDCS IN                        \ Is it still low or has it been pulled up (by the card's pullup)
    &SDCS OUTSET                \ force back to an inactive CE output
    ;

\ Fetch a byte from the SD (clock in 1's)
pub SD@ ( -- 8b )            
    -1 RUNMOD >B
    ;
\ Write a byte to the SD
pub SD! ( 8b -- )
    #24 SHL RUNMOD DROP
    ;
pub SD2@ ( -- word )    
     SD@ 8 SHL SD@ OR
     ;
pub SD4@ ( -- long )
     BL @CNT COGREG! -1 RUNMOD 8 @CNT COGREG!
     ;        
\ Faster byte wide clocks (8/count)
pub SDCLK ( cnt -- )    
    FOR SD@ DROP NEXT
    ; 

#10,000 == &sdto                    \ SD timout period - can be very long

\ do what needs to be done if the SD card is busy (LED etc) - can be rewritten
pri SDBUSY ( state -- )        sdbusy W@ ?DUP IF JUMP ELSE DROP THEN ;

\ Initialize the SD I/O and basic card
pub SDPINS ( ce-miso-mosi-clk -- )
    sdpins !                \ remember which pins were used for BACKUP
pub !SDIO 
     sdpins @
    DUP #24 SHR MASK ' &SDCS 1+ !
    MODPINS
    ucard W~
    sdbusy W~

    [SPIO]
    1 ms
    &SDCS OUTCLR                \ select the card
    @MOSI COGREG@ OUTSET            \ Set MOSI high
     @SCK COGREG@ OUTCLR            \ clock low (and as an output)
     #100 FOR CLOCK CLOCK NEXT        \ toggle clock high and low nice and slow initially
    &SDCS OUTSET
    16 SDCLK                    \ 16 byte-wide clocking cycles 
    &SDCS OUTCLR
    16 SDCLK
    &SDCS OUTSET
    ;

( SD CARD RESPONSES )



( Response fetch )
pub RES@ ( -- res ) \ Continue reading until a response is received (with timeout)
    $FF &sdto 0                    \ TIMEOUT PERIOD
      DO SD@ DUP $FF <>
        IF NIP LEAVE ELSE DROP THEN
      LOOP
    ;


( Issue a command to the SD card )
pub CMD    ( data cmd -- res )     \ Send the command to the SD card and read result
    [SPIO]
     sdpins @ MODPINS            \ Update the RUNMOD pin masks in case another module has been used
    SD@ DROP                \ extra clocks
     &SDCS OUTCLR            \ Select the card
    $3F AND $40 OR SD!        \ send the command
    BL @CNT COGREG!            \ 32-bit transfer  
     [SPIO] RUNMOD DROP        \ send data
    8 @CNT COGREG!             \ back to 8-bit transfers
    crc C@ SD!                \ dummy checksum ( is valid for initializing) 
    RES@                    \ fetch the resPONSE
    ;

\ Send an ACMD to the card and return with response
pub ACMD ( data acmd -- res )     
     0 #55 CMD DROP CMD
     ;
pri STAT@ ( -- stat )
     0 #13 CMD RES@ 8 SHL OR
     ;

pri SDERR? ( -- flg ; return SD bit flag errors)
    0 CARD? 0= 1 AND OR
    sdrd @ 0= 2 AND OR
    sdwr @ 0= 4 AND OR
    ocr @ 0= 8 AND OR
    ;
pri MARKER? ( marker -- flg ; Find SD marker and return true before timeout )
    &sdto BEGIN OVER SD@ <> WHILE 1- DUP 0= IF NIP EXIT THEN REPEAT
    2DROP TRUE
    ;
pri SDDAT! ( adr -- ; Wait for read token and read SD data into buffer )
    =dtk MARKER? IF 16 ADO SD@ I C! LOOP 3 SDCLK ELSE DROP THEN
    ;



\ In SPI Mode, only the OCR, CSD and CID registers are accessible. 
pri ?SDTIMEOUT        7 TIMEOUT? IF FALSE R> DROP EXIT THEN ;

\ Initialise the SD card (with timeout)
pub !SD ( -- ocr|false ; Initialise the SD card in SPI mode and return with the OCR )
    ocr ~
    cid 16 ERASE csd 16 ERASE
     0 FILE
    SDBUF BLKSIZ 2* 2* ERASE
     sector 16 ERASE
    sdpins @ SDPINS                \ reload SDPINS in case this is a reboot
    CARD? 
    IF
      #1000 7 TIMEOUT
      $95 crc C!
        BEGIN ?SDTIMEOUT 0 0 CMD 1 = UNTIL
      $87 crc C!    
           BEGIN ?SDTIMEOUT $1AA 8 CMD 1 = UNTIL
       SD4@ $1AA =
      IF  
          BEGIN ?SDTIMEOUT $4000.0000 #41 ACMD 0= UNTIL
          BEGIN ?SDTIMEOUT 0 #58 CMD 0= UNTIL
         SD4@ DUP ocr !
        DUP 
         IF
           0 #10 CMD 0= IF cid SDDAT! THEN
           0 9 CMD 0= IF csd SDDAT! THEN
         THEN
      ELSE
        FALSE
      THEN
    ELSE
      FALSE
    THEN
    &SDCS OUTSET
    ;

\ These functions are only for reporting to the console
pri XSHR
    0 DO
    0 extbuf 16
      ADO I C@ DUP 2/ ROT OR I C!
        1 AND IF $80 ELSE 0 THEN
      LOOP
     DROP
     LOOP
    ;
\ Read card information fields
pri INFO@    ( bith bitl adr -- dat )
    extbuf 16 CMOVE 
     DUP XSHR - 1+ MASK
    1-
    0 extbuf $0C + 4 ADO 8 SHL I C@ OR LOOP
    AND
    ;

TABLE tval    
    #00 | #10 | #12 | #13 | 
      #15 | #20 | #25 | #30 |
    #35 | #40 | #45 | #50 | 
     #55 | #60 | #70 | #80 |
    
\ 0=0.5 mA, 1=1 mA, 2=5 mA, 3=10 mA, 4=25 mA, 5=35 mA, 6=60 mA, 7=100mA
\ 0=1 mA, 1=5 mA, 2=10 mA, 3=25 mA, 4=35 mA, 5=45 mA, 6=80 mA, 7=200
TABLE ma
    #0 | #1 | #5 | #10 | #25 | #35 | #60 | #100 |
    #1 | #5 | #10 | #25 | #35 | #45 | #80 | #200 |

\ Card information
pub .CARD
    sdhc C~
    #20 .HEAD                        \ set head indent into cid
    \    0123456789ABCDEF012345
    " MFG" .HEAD     cid C@ .BYTE
    " OEM" .HEAD    cid 1+ 2 TYPE
    " PID" .HEAD    cid 3 + 5 TYPE
    " REV" .HEAD    cid 8 + C@ .BYTE
    " S/N" .HEAD      cid 9 + U@ .LONG        \ serial #
    " DATE" .HEAD    cid #13 + C@
    8 SHL         cid #14 + C@ OR
    DUP 4 SHR #2000 + #10 .NUM
    ." /" $0F AND #10 .NUM
    " CRC" .HEAD    cid #15 + C@ 2/ .BYTE

    "   CSD REGISTER" .HEAD
    #127 #126 csd INFO@
    IF CR ." *SDHC CARD* " sdhc C~~ THEN

    " TACC" .HEAD         #119 #112 csd INFO@
    DUP 3 SHR tval + C@ ( *10)
    SWAP 3 AND 1 SWAP 1+ 0 DO #10 * LOOP #100 / * .DECX ." ns"
    " NSAC" .HEAD        #111 #104 csd INFO@ 
     DUP 3 SHR tval + C@ #10,000 ROT 1+ 0 DO #10 * LOOP * .DECX ." bps" 

    " TRANS" .HEAD        #103 #96 csd INFO@ .DECX ." MHz"
     " BLKLEN" .HEAD        #83 #80 csd INFO@ DUP blklen C! |< .DECX
    sdhc C@ 
     IF
       " SIZE" .HEAD        #69 #48 csd INFO@
    ELSE
      " SIZE" .HEAD        #73 #62 csd INFO@ 
       #49 #47 csd INFO@ 2+ SHL #10 SHR
     THEN
    #10 blklen C@ - SHR   DUP sdsize ! .DECX ." MB"
{
    sdhc C@ NOT
     IF
      " Read Current min" .HEAD    #61 #59 csd INFO@ ma + C@ .DECX ." ma"
      " Read Current max" .HEAD    #58 #56 csd INFO@ 8 + ma + C@ .DECX ." ma"
       " Write Current min" .HEAD    #55 #53 csd INFO@ ma + C@ .DECX ." ma"
      " Write Current max" .HEAD    #52 #50 csd INFO@ 8 + ma + C@ .DECX ." ma"
    THEN
}
    CR "_" BL EMITS
    ;


{ *** BLOCK BUFFERS *** }

\ Load PASM I/O module into kernel - ensure pin masks are set by forcing an update (if sharing RUNMODs)
pri (SDWR) ( src cnt -- )        ( sdpins @ MODPINS ) [SDWR] RUNMOD ;    
pri (SDRD) ( dst cnt -- )        ( sdpins @ MODPINS ) [SDRD] RUNMOD ;
    
pub SDRDBLK ( dst -- crc )
    BLKSIZ (SDRD)
    [SPIO]
    SD2@                    \ read crc
    31d MASK OR                \ force crc as TRUE flag (msb set)
    ;


pri PROCESS_TOKEN    ( dst token -- flg )
     =dtk = IF SDRDBLK EXIT THEN
    2DROP                        \ drop token and dst
    2 SDCLK STAT@ DROP            \ discard cycles
    &SDCS OUTSET
    FALSE
    ;

\ SD CARD READ BLOCK - measured 2.3ms to 6ms on 4GB Sandisk
pub SDRD    ( sector dst -- crc | false ; Read sector from SD into dst )
    ON SDBUSY
    SWAP #17 CMD                \ read block command
      IF
      DROP FALSE
      ELSE
      RES@ PROCESS_TOKEN            \ process read token
       THEN
    &SDCS OUTSET
    OFF SDBUSY
    DUP sdrd !                   \  save crc/flg
    ;

\ Write to a sector in the SD card - measured 2.84ms on 4GB SanDisk
pub SDWR    ( src sect -- flg ; Write from src to xdst in the SD )
    ON SDBUSY
    &SDCS OUTSET 1 SDCLK 
     &SDCS OUTCLR 3 SDCLK
    #24 CMD ( src res )
    0=
      IF
         3 SDCLK =dtk SD!
        BLKSIZ (SDWR)                \ Write sector
        [SPIO]                    \ switch back to normal SPI mode
        0 MARKER? $FF MARKER? AND
        ELSE FALSE
      THEN
    &SDCS OUTSET
    OFF SDBUSY
    DUP sdwr !
    [SPIO]
    ;


\ Quick terminal command to initialise a card and dislay some information
pub SD            
     sdpins @ SDPINS
     CARD? IF !SD IF    CR ." Card inserted " .CARD CR ELSE ." Card Error " THEN
     ELSE ." No Card " THEN
     ;
\ Poll to check for insertion or removal
\ To activate either use KEYPOLL SDPOLL or incorporate SDPOLL into another polling task
pub SDPOLL
     card? C@ 0= CARD? DUP card? C! AND IF SD THEN
     ;


 

\ X words - denotes virtual memory words which in this case uses SD memory.

\ 192.5ms to save 32K of memory
pub XSAVE ( src xdst cnt -- )
      SWAP 9 SHR SWAP 0 DO
      2DUP SDWR DROP
      1+ SWAP BLKSIZ + SWAP
      BLKSIZ +LOOP
    2DROP
    ; 

pub XLOAD ( xsrc dst cnt -- )
      ROT 9 SHR ROT ROT 0 DO
      2DUP SDRD DROP
      BLKSIZ + SWAP 1+ SWAP
      BLKSIZ +LOOP
    2DROP
    ; 


( Virtual memory addressing is suitable for addressing the first 4GB of the memory )

\ load a sector into the SDBUF - takes around 1.9ms
pub SECTOR ( sect -- )
     DUP sector @FILE ! SDBUF SDRD sectcrc @FILE !
     ;
pub ACTIVE? ( -- flg )          
     sectcrc @FILE @ $80000000 <>
     ;
wrflg C~

pub @WRFLG ( -- mask caddr   \ write flag bit address - use SET CLR or SET? ) 
     filesel C@ MASK wrflg
     ;

pub WRSECT ( write the current sector back to the storage media )
     SDBUF sector @FILE @ SDWR 
    ;
\ Write the sector buffer if it has been modified
pub FLUSH
     @WRFLG SET? 
     IF WRSECT @WRFLG CLR THEN
    ;

\ Translate a virtual address into a HUB RAM address - may have to read a sector
pub XADR ( xaddr -- addr )                \ make sure sector is loaded into SDBUF and return with physical address
    >L
    IX 9 SHR sector @FILE @ <>            \ buffered already?
    IF                            \ No
       FLUSH
      IX 9 SHR SECTOR                 \ get sector into SDBUF if not already there
    THEN
    L> BLKSIZ 1- AND SDBUF +            \ return with physical address
    ;
pub XC@ ( xaddr -- byte )
    XADR C@
    ;
pub XC! ( byte xaddr -- )
    XADR C!  
     @WRFLG SET  
    ;


( Dump virtual memory )
pub XDUMP ( vadr cnt -- )
    ADO
      I XADR DROP
      I   IX I - BLKSIZ MIN ( vadr size ) 
       SDBUF I BLKSIZ 1- AND + BDUMP                \ dump the buffer (or partial)            
      KEY? IF $1B = IF LEAVE ELSE KEY DROP THEN ELSE DROP THEN
      BLKSIZ
    +LOOP
    ;



\ **********************************************************

{    
\ Terminal background polling method to check for card insertion
KEYPOLL SDPOLL

\ Background timer method
\ Perform SDPOLL and reload timer
pub SDPOLL2   SDPOLL #200 1 TIMEOUT ;
\ Set the ALARM action for SDPOLL2 and trigger it
' SDPOLL2 1 ALARM SDPOLL2
}

{ Autorun method V2.0 120917 modified
pub SDAUTO    !SD ' SDPOLL keypoll W! ;
AUTORUN SDAUTO  
}



( TARGET PLATFORM PIN DEFINITIONS ) 
{
Target platforms can be selected by name or defined if necessary. The long form of pin definition may be
used (SDCS SCK MOSI MISO) ( ce&miso&mosi&clk -- ) first after which the variable sdpins may be read and a single long used
for the runtime definition.  
}
{
pub P1144        &24.27.26.25 SDPINS ;
pub P1107        #P26 PINCLR      &25.27.29.28 SDPINS ;
pub P1145        #P26 PINCLR      &19.18.29.28 SDPINS ;
pub CE0077        &24.27.26.25 SDPINS ;
pub CE0972        &25.05.29.28 SDPINS ;
pub CE1200        &21.24.23.22 SDPINS ;
pub CE1372        &26.27.29.28 SDPINS ;

\ PARALLAX
pub BOECARD? ( -- flg )
     @MOSI COGREG@ DUP INPUTS    \ float MOST line with card-detect pullup/down
     8 us
     DUP IN 0=                \ card inserted if low
     SWAP OUTCLR
     ; 

pub BOE            &25.22.24.23 SDPINS ' BOECARD? ucard W! ;
pub PPUSB            &03.00.02.01 SDPINS ;
pub SPINNERET        &19.16.20.21 SDPINS #P17 2 MASKS INPUTS ;

}

{


pub XMAP ( src cnt -- )
    BEGIN
      CR OVER .LONG ." : "
      OVER BLKSIZ 5 SHL    \ 32 SECTORS / LINE 
       ADO
        0 I XADR BLKSIZ ADO I C@ + LOOP 9 SHR .MAP SPACE
        BLKSIZ 
       +LOOP
    BLKSIZ 5 SHL ROT OVER + ( cnt inc src+inc )
     ROT ROT - DUP 0= 
     UNTIL
     2DROP
    ;




}

]~   \ end of load symbols
END



