FORGET W5200.fth
TACHYON
[~

\ Module name - reports in list of modules using MODULES word.
pub W5200.fth        ." WIZNET W5200 driver 131107.0000 " ;

IFNDEF SPIRD
."           !!! Needs V2.3 kernel or greater !!! "
!!!
}
{                  *** WIZnet W5100/5200 driver ***

Implements the SPI interface to the WIZnet W5100 including the primitives to talk to the chip 
The aim is to make this talk to the W5200 as well but the W5100 on the Spinneret is being used for development

130218
Looking at dedicating a cog just to high-speed SPI access which will not only speed up processing but also make it easier to multi-task. I/O will be via simple hub variables.


NOTE: Refresh webpage for latest version
}



( HARDWARE DEFINITIONS )
IFNDEF &WNCS
#P12    |< == &WNDO            '' MISO from WIZNET
#P13    |< == &WNCS            '' WIZNET CS
#P14    |< == &WNCK
#P15    |< == &WNDI            '' MOSI to WIZNET
}
\ Compile time constant for pin configuration
&WNCS >| #24 SHL &WNDO >| 16 SHL + &WNDI >| 8 SHL + &WNCK >| + == wizpins

\ The interrupt signal can be read from MOSI by floating MOSI and reading the state
\ Only 4 Propeller I/O are required to interface to the WIZnet W5200 rather than 7 I/O
\ Additional hardware as used on CE1372 for reset and power-down



\ Diagram: WIZnet SPI timing 

{
3. Write target address for transmission on SPDR register (SPI Data Register).
4. Write OP code and data length for transmission on SPDR register.
5. Write desired data for transmission on SPDR register.
6. Configure nSCS as ‘Low’ (data transfer start)
7. Wait for reception complete
8. If all data transmission ends, configure nSCS as ‘High’

}


pub LC! ( byte xadr -- )
    SWAP #24 SHL SWAP 1 
\ Write to the WIZnet register via SPI
pub WIZX! ( data address cnt -- )
    &WNCS OUTCLR 
    SWAP 16 SHL SPIWR SPIWR DROP                    \ write address
     DUP $8000 OR 16 SHL SPIWR SPIWR DROP ( data cnt )    \ output count
     FOR SPIWR NEXT DROP
     &WNCS OUTSET                            \ chip select, write 32-bits SPI, chip deselect
    ;
pub LW! ( word xadr -- )
     SWAP 16 SHL SWAP 2 WIZX!
     ;
pub L! ( long xadr -- )
     4 WIZX!
;


\ Read from the WIZnet register via SPI
pub LC@ ( address -- data )
    1
pri WIZX@ ( address cnt -- data )
    &WNCS OUTCLR
    SWAP 16 SHL SPIWR SPIWR DROP        \ send addr
     DUP 16 SHL SPIWR SPIWR DROP        \ send count
    0 SWAP FOR SPIRD NEXT             \ read back data
     &WNCS OUTSET 
    ;
{
    BL @SPICNT!                    \ 16-bit address + R/W + 15-bit length
    &WNCS OUTCLR
    SWAP 16 SHL OVER OR SPI DROP        \ send addr+op+cnt
    2* 2* 2* @SPICNT! 0 SPI            \ read back data
     &WNCS OUTSET 
    ;
}
\ Read a word from the WIZnet register
pri LW@ ( reg -- word )
    2 WIZX@ 
    ; 
\ Read a long from the WIZnet register
pri L@ ( reg -- long )
    4 WIZX@
    ;

{ *** W5200 MEMORY MAP ***
0000-0036    Common Registers
4000-4FFF    Socket Registers
8000-BFFF    Transmit memory
C000-FFFF    Receive memory
}

\ Read a block of memory from the WIZnet
pri LREAD ( src dst cnt -- )
    ROT ( dst cnt src )                \ 32-bit transfers
     &WNCS OUTCLR
    16 SHL SPIWR SPIWR DROP ( dst cnt)        \ transmit src address 
     DUP 16 SHL SPIWR SPIWR DROP ( dst cnt )    \ transmit count
    ADO 0 SPIRD I C! LOOP
     &WNCS OUTSET 
    ;

\ Write a block to the WIZnet - 3.2ms/512 bytes = = 6.25us/byte = 160kB/sec
pri LWRITE ( src dst cnt -- )
    SWAP
     &WNCS OUTCLR
    16 SHL SPIWR SPIWR DROP                \ Send address
     DUP $8000 + 16 SHL SPIWR SPIWR DROP     \ send total count
    ( src cnt ) ADO I C@ SPIWRB DROP LOOP    \ now just send the bytes sequentially
     &WNCS OUTSET 
    ;


( SOCKET INTERRUPTS )
4    |< ==     &SENDOK
3    |< ==     &TIMEOUT
2    |< ==     &RECV
1    |< ==     &DISCON
0    |< ==     &CON
{
( SOCKET STATUS CODES )
$00    ==    SOCK_CLOSED
$13    ==    SOCK_INIT 
$14    ==    SOCK_LISTEN
$17    ==    SOCK_ESTABLISHED
$1C    ==    SOCK_CLOSE_WAIT
$22    ==    SOCK_UDP
$32    ==    SOCK_IPRAW
$42    ==    SOCK_MACRAW
$5F    ==    SOCK_PPOE

( SOCKET TRANSIENT STATUS CODES )
$15    ==    SOCK_SYNSENT
$16    ==    SOCK_SYNRECV
$18    ==    SOCK_FIN_WAIT
$1A    ==    SOCK_CLOSING
$1B    ==    SOCK_TIME_WAIT
$1D    ==    SOCK_LAST_ACK
$11    ==    SOCK_ARP
$21    ==    SOCK_ARP1
$31    ==    SOCK_ARP2
}

\ current socket register address offset
$4000 == #SOCKET        
WORD socket
\ Set the socket number ( calculates address and sets as a SOCKET constant )
pub SKT
pub SOCKET ( socket -- )    DUP socket C! $40 + 8 SHL ' #SOCKET 1+ ! ;

pub @SOCKET        #SOCKET + ;

\ Socket registers ( -- addr )
pub sMODE        0 @SOCKET ;        \ mode
pub sCMD!        1 @SOCKET LC! ;    \ command
pub sINTS        2 @SOCKET ;        \ interrupts
pub sSTAT        3 @SOCKET ;        \ status 
pub sPORT        4 @SOCKET ;        \ 2 byte src port
pub sDHAR        6 @SOCKET ;        \ 6 byte dest hardware address
pub sDIP        $0C @SOCKET ;    \ 4 byte dest IP address
pub sDPORT        $10 @SOCKET ;    \ 2 byte dest port
pub sSSIZE        $12 @SOCKET ;    \ 2 byte dest max seg size
pub sPRO        $14 @SOCKET ;    \ protocol in IP raw mode

pub sRXMEM        $1E @SOCKET ;
pub sTXMEM        $1F @SOCKET ;
pub TXFREE@        $20 @SOCKET LW@ ;
pub TXREAD        $22 @SOCKET ;
pub TXWRITE        $24 @SOCKET ;
pub RXSIZE@        $26 @SOCKET LW@ ;
pub RXREAD        $28 @SOCKET ;
pub RXWRITE        $2A @SOCKET ;

\ Protocol modes
pub CLOSED                0 sMODE LC! ;
pub TCP                1 sMODE LC! ;
pub UDP                2 sMODE LC! ;
pub IPRAW                3 sMODE LC! ;
pub MACRAW                4 sMODE LC! ;
pub PPPoE                5 sMODE LC! ;

$8000        == TXBASE0
$07FF        == TXMASK0
$C000        == RXBASE0
$07FF        == RXMASK0

\ common registers
LONG srcip
\ Access common registers
pub wMODE ( mask -- )        0 LC! ;
pub GATEWAY ( addr -- )        1 L! ;
pub SUBNET ( mask -- )        5 L! ;
pub SHAR ( high low  -- )    $0B L! 9 LW! ;
pub SRCIP ( long -- )        DUP srcip ! $0F L! ;

pub INTS@ ( -- ints )        $15 LC@ ;
pub INTMASK ( mask -- )        $16 LC! ;
pub RETRYTIME ( val -- )    $17 LW! ;
pub RETRYS ( cnt -- )        $19 LC! ;
pub RXSIZ ( mask -- )        $1A LC! ; \ default sets 2K for each socket (%01010101)
pub TXSIZ ( mask -- )        $1B LC! ;

pub UIP ( ip -- )            $2A L! ;
pub UPORT ( port -- )        $2E LW! ;




( SOCKET COMMANDS - use currently select socket )
pub sOPEN          1 sCMD! ; 
pub sLISTEN          2 sCMD! ;
pub sCONNECT      4 sCMD! ;
pub sDISCON          8 sCMD! ;
pub sCLOSE        $10 sCMD! ;
pub sSEND        $20 sCMD! ;
pub sSENDMAC    $21 sCMD! ;
pub sSENDKEEP    $22 sCMD! ;
pub sRECV        $40 sCMD! ;

pub PORT! ( srcport -- )        sPORT LW! ;





RND             == mac        \ compile time constant
pub MAC        $00.08 mac ;
&192.168.16.1    == gateway
&192.168.16.150    == srcip
&255.255.255.0    == subnet
 
\ Init the SPI for the WIZnet chip
pub !WIZIO
    !PCB
    wizpins $00FFFFFF AND SPIPINS
    &WNCS OUTSET                \ Chip select is an output - active low
    &WNCK OUTCLR                \ Clock is an output - leave low
    &WNDI OUTCLR                \ MOSI is an output
    ON WRESET OFF WPWRDN OFF WRESET
;

pub !WIZIP
    gateway GATEWAY            \ Assign a default gateway
    subnet SUBNET
    srcip SRCIP
    MAC SHAR                    \ random MAC at compile time, Need to have a unique ID    
    ;
}
{ Since the W5200 read and write index registers are unreadable until a connection has been established
or cannot be written incrementally until a send then these are buffered when they become readable
} 

WORD txwr    #14 ALLOT
pub @txwr   txwr socket C@ 2* + ;

{
pub BLKSEND ( addr -- )                    \ send one complete sector at buffer address
    @txwr W@ -1 = IF TXWRITE LW@ @txwr W! THEN    \ force an update (assumed valid) if the index is "invalid"
     @txwr W@ $7FF AND                        \ mask it according to buffer size (default 2K)
    TXBASE0 socket C@ #11 SHL + + ( src dst )        \ write pointer
     BLKSIZ LWRITE ( src dst cnt ) 
    BLKSIZ @txwr W+!
    @txwr W@ TXWRITE LW! sSEND
    ;
}

BYTE autosend
autosend C~~
pub LANEMIT ( ch -- )
    @txwr W@ -1 = IF TXWRITE LW@ @txwr W! THEN    \ force an update (assumed valid) if the index is "invalid"
     @txwr W@ $7FF AND                        \ mask it according to buffer size (default 2K)
    TXBASE0 socket C@ #11 SHL + +                \ write pointer
    OVER SWAP LC!
      @txwr W++
    $0A = autosend C@ AND IF @txwr W@ TXWRITE LW! sSEND THEN
    ;

pub LANKEY ( -- ch )
     RXSIZE@
       IF RXREAD LW@ RXBASE0 socket C@ #11 SHL + OVER $7FF AND + LC@
         SWAP 1+ RXREAD LW! sRECV
       ELSE keypoll W@ ?DUP IF CALL THEN
         0
       THEN 
    ;

\ Redirect console output to the LAN
pub LAN     
    ' LANEMIT uemit W! 
     ' LANKEY ukey W! 
    ;




\
\ Diagnostics - only needed for debugging

pub WDUMP ( addr cnt -- )
    DUP ROT ROT          \ maintain "cnt"
     ADO I BUFFERS 3RD $100 MIN LREAD
      I OVER $100 MIN BUFFERS BDUMP DUP $100 MIN 
     +LOOP
    DROP
    ;

pub @BUFFER        BUFFERS + ;

pri .IPX    DUP >B $30A .NUM ." ." 8 SHR ;
pub .IP ( off -- )
     DECIMAL 3 SPACES 
pub .IP1
    @BUFFER U@ 
    .IPX .IPX .IPX .IPX DROP
    ;

pub .SSTAT ( byte -- )
     SWITCH SPACE
       0 CASE ." CLOSED" BREAK
     $13 CASE ." INIT" BREAK
     $14 CASE ." LISTEN " BREAK
     $17 CASE ." ESTABLISHED "  $0C .IP BREAK
     $1C CASE ." CLOSED WAIT" BREAK
     $22 CASE ." UDP OPEN" BREAK
     $32 CASE ." IPRAW OPEN" BREAK
     ;

" CLOSEDTCP   UDP   IPRAW MACRAWPPPoE !!06!!!!07!!" 0 STRING MODE$

pub .PTR          I + C@++ 8 SHL SWAP C@ + ( $7FF AND ) .WORD "." EMIT ;
IFNDEF SWAPB
pub SWAPB        DUP 8 SHR SWAP >B 8 SHL OR ;
}
pub W
pub WIZ?
    CR ." ************ W5200 STATUS ************ "
     0 BUFFERS $40 LREAD    \ read in common registers
    CR ." LINK " 5 MASK BUFFERS $35 + SET? IF ." *UP*" ELSE ." DOWN" THEN  
    CR ." CHIP VER  " $1F @BUFFER C@ .DEC
    CR ." SRC IP "    $0F .IP
     CR ." MASK   "    5 .IP
     CR ." GATEWAY"    1 .IP
    CR ." MAC       "    BUFFERS 9 + 6 ADO I C@ .BYTE ." ." LOOP
pub SOCKETS?
    CR ." *** SOCKETS *** "
      $4000 BUFFERS $800 LREAD    \ read in all socket registers    
     BUFFERS $800 ADO 
       CR I BUFFERS - 8 SHR 7 AND . ." : " 
       I C@ ." MODE=" SPACE 6 * MODE$ + 6 TYPE 
       ."  PORT#" 4 I + W@ SWAPB .DEC SPACE $10 I + W@ SWAPB .DEC  TAB
      ."  TXRW=" $22 .PTR $24 .PTR 
       ." RXRW=" $28 .PTR $2A .PTR
       ." RXSZ=" $26 .PTR
        ." IR=" 2 I + C@ .BYTE
       TAB ."  STAT=" 
       3 I + C@ DUP .BYTE .SSTAT
     $100 +LOOP 
     CR
    ;

pri WDD    OVER CR ." Socket# " . DUP $30 WDUMP $100 + SWAP 1+ SWAP  ;
pub WD
     ."  Dumping WIZNET registers "
    CR ." Common registers "
    0 $40 WDUMP
    0 $4000 WDD WDD WDD WDD WDD WDD WDD WDD 2DROP
    ;









]~
END


